#include <algorithm>
#include <cstring>
#include <iostream>
#include <vector>
#include "abgame.h"
#include "answer.h"
#include "questiongen.h"
#include "timer.h"

typedef struct {
  int question;
  int answer;
} upper_bound_function_t;

#if FUNCTION >= 201 && FUNCTION <= 299
  #include "upper_bound_2_pegs.h"
#elif FUNCTION >= 301 && FUNCTION <= 399
  #include "upper_bound_3_pegs.h"
#elif FUNCTION >= 4001 && FUNCTION <= 4999
  #include "upper_bound_4_pegs.h"
#else
  static const upper_bound_function_t upper_bound_function[1][1] = {{{0, 0}}};
#endif

using namespace std;

static ostream & operator<<(ostream &os, Answer const &ans) {
  os << '(' << ans.B() << ',' << ans.W() << ')';
  return os;
}

static ostream & operator<<(ostream &os, Configuration const &conf) {
  for (int i = 0; i < pegs; ++i)
    // os << (char)(conf[i] + 'a');
    os << (int)conf[i] << ' ';
  return os;
}

static ostream & operator<<(ostream &os, Maset const &maset) {
  for (size_t i = 0; i < maset.size(); ++i)
    os << maset[i] << '\n';
  return os;
}

/** Masets **/

class AllABgameConfigurations : public Maset {
  public:
    AllABgameConfigurations(int additional_colors) {
      Peg peg_tbl[colors + additional_colors];
      m.clear();
      Init(peg_tbl, additional_colors);
      m.push_back(Configuration(peg_tbl));
      while (Next(peg_tbl, additional_colors))
        m.push_back(Configuration(peg_tbl));
    }

  private:
    void Init(Peg *peg_tbl, int additional_colors) {
      for (int i = 0; i < colors; ++i)
        peg_tbl[i] = i;
      for (int i = colors; i < colors + additional_colors; ++i)
        peg_tbl[i] = colors; // An additional color
    }

    bool Next(Peg *peg_tbl, int additional_colors) {
      int i, j, k, l, tmp;
      if (peg_tbl[pegs - 1] < peg_tbl[colors + additional_colors - 1]) {
        for (j = pegs; j < colors + additional_colors; ++j) {
          if (peg_tbl[j] > peg_tbl[pegs - 1]) {
            tmp = peg_tbl[pegs - 1];
            peg_tbl[pegs - 1] = peg_tbl[j];
            peg_tbl[j] = tmp;
            return true;
          }
        }
      }
      else {
        for (i = pegs, k = colors + additional_colors - 1; i < k; ++i, --k) {
          tmp = peg_tbl[i];
          peg_tbl[i] = peg_tbl[k];
          peg_tbl[k] = tmp;
        }
        j = pegs - 2;
        while (j >= 0 && peg_tbl[j] >= peg_tbl[j + 1])
          --j;
        if (j < 0)
          return false;
        l = colors + additional_colors - 1;
        while (peg_tbl[j] >= peg_tbl[l])
          --l;
        tmp = peg_tbl[j];
        peg_tbl[j] = peg_tbl[l];
        peg_tbl[l] = tmp;
        for (i = j + 1, k = colors + additional_colors - 1; i < k; ++i, --k) {
          tmp = peg_tbl[i];
          peg_tbl[i] = peg_tbl[k];
          peg_tbl[k] = tmp;
        }
      }
      return true;
    }
};

class ABgameFixedQuestion : public Maset {
  public:
    ABgameFixedQuestion(int i) {
      Peg peg_tbl[pegs];
      if (FUNCTION == 0) {
        for (int j = 0; j < pegs; ++j)
          peg_tbl[j] = j + i * pegs;
      }
      else {
        cerr << "ERROR: Wrong parameter combination in ABgameFixedQuestion" << endl;
        exit(1);
      }
      m.push_back(Configuration(peg_tbl));
    }
};

class ABBgameFixedQuestion : public Maset {
  public:
    ABBgameFixedQuestion(int i) {
      Peg peg_tbl[pegs];
      if (FUNCTION >= 201 && FUNCTION <= 299)
        i = upper_bound_function[FUNCTION - 201][i].question;
      else if (FUNCTION >= 301 && FUNCTION <= 399)
        i = upper_bound_function[FUNCTION - 301][i].question;
      else if (FUNCTION >= 4001 && FUNCTION <= 4999)
        i = upper_bound_function[FUNCTION - 4001][i].question;

      for (int j = 0; j < pegs; ++j) {
        int k = i + j;
        if (k < 0)
          peg_tbl[j] = k + colors;
        else if (k >= colors)
          peg_tbl[j] = k - colors;
        else
          peg_tbl[j] = k;
      }
      m.push_back(Configuration(peg_tbl));
    }
};

class ABBgameSemiFixedQuestions : public Maset {
  public:
    ABBgameSemiFixedQuestions() {
      Peg peg_tbl[pegs];
      for (int i = 0; i < colors; ++i) {
        for (int j = 0; j < pegs; ++j) {
          int k = i + j;
          peg_tbl[j] = k >= colors ? k - colors : k;
        }
        m.push_back(Configuration(peg_tbl));
      }
    }
};

class AllMastermindConfigurations : public Maset {
  public:
    AllMastermindConfigurations(int additional_colors) {
      Peg peg_tbl[pegs];
      m.clear();
      Init(peg_tbl, additional_colors);
      do
        m.push_back(Configuration(peg_tbl));
      while (Next(peg_tbl, additional_colors));
    }

  private:
    void Init(Peg *peg_tbl, int additional_colors) {
      for (int i = 0; i < pegs; ++i)
        peg_tbl[i] = 0;
    }

    bool Next(Peg *peg_tbl, int additional_colors) {
      for (int i = 0; i < pegs; ++i) {
        if (peg_tbl[i] != colors + additional_colors - 1) {
          for (int j = 0; j < i; ++j)
            peg_tbl[j] = 0;
          ++peg_tbl[i];
          return true;
        }
      }
      return false;
    }
};

class AllABBendgameConfigurations : public Maset {
  public:
    AllABBendgameConfigurations(vector<Peg> const vv[pegs]) {
      Peg peg_tbl[pegs];
      int idx[pegs];
      int size[pegs];
      m.clear();
      for (int i = 0; i < pegs; ++i) {
        idx[i] = 0;
        size[i] = vv[i].size();
      }
      do
        if (Valid(peg_tbl, idx, vv))
          m.push_back(Configuration(peg_tbl));
      while (Next(idx, size));
    }

  private:
    bool Next(int *idx, int *size) {
      for (int i = 0; i < pegs; ++i) {
        if (idx[i] != size[i] - 1) {
          for (int j = 0; j < i; ++j)
            idx[j] = 0;
          ++idx[i];
          return true;
        }
      }
      return false;
    }

    bool Valid(Peg *peg_tbl, int const *idx, vector<Peg> const *vv) {
      for (int i = 0; i < pegs; ++i)
        peg_tbl[i] = vv[i][idx[i]];
      for (int i = 0; i < pegs; ++i)
        for (int j = i + 1; j < pegs; ++j)
          if (peg_tbl[i] == peg_tbl[j])
            return false;
      return true;
    }
};

/** Asking questions and computing answers **/

class AnsweredMaset {
  public:
    AnsweredMaset(Maset const &mas, Configuration const &question) {
      asked_question = question;
      masets.resize(possible_answers);
      for (int i = 0; i < possible_answers; ++i)
        masets[i] = make_pair(Maset(), Answer(i));
      for (size_t i = 0; i < mas.size(); ++i)
        masets[Answer(mas[i], question).Code()].first.push(mas[i]);
      sort(masets.begin(), masets.end(), CompareMasets);
      if (atradeoff > 0 && masets.size() > atradeoff) {
        masets.erase(masets.begin() + atradeoff, masets.end());
      }
    }

    size_t size() const {
      return masets.size();
    }

    size_t max_size() const {
      return masets[0].first.size();
    }

    Maset const & GetMaset(size_t i) const {
      return masets[i].first;
    }

    Answer const & GetAnswer(size_t i) const {
      return masets[i].second;
    }

    Configuration const & GetAskedQuestion() const {
      return asked_question;
    }

    void swap(AnsweredMaset &a) {
      masets.swap(a.masets);
      asked_question.swap(a.asked_question);
    }

  private:
    typedef pair<Maset, Answer> MasetWithAnswer;

    vector<MasetWithAnswer> masets;
    Configuration asked_question;

    // Masets are sorted in the decreasing order of their size.
    static bool CompareMasets(MasetWithAnswer const &p1, MasetWithAnswer const &p2) {
      return p1.first.size() > p2.first.size();
    }
};

static void swap(AnsweredMaset &left, AnsweredMaset &right) {
  left.swap(right);
}

static ostream & operator<<(ostream &os, AnsweredMaset const &ans_mas) {
  cout << "Present question " << ans_mas.GetAskedQuestion() << '\n';
  for (size_t i = 0; i < ans_mas.size(); ++i) {
    Maset const & mas = ans_mas.GetMaset(i);
    if (mas.size())
      cout << "\nThe answer "
           << ans_mas.GetAnswer(i)
           << " gives the maset\n"
           << mas;
  }
  cout << '\n';
  return os;
}

class AllAnswers {
  public:
    AllAnswers(Maset const &maset, Maset const &questions, size_t upper_size) {
      for (size_t i = 0; i < questions.size(); ++i) {
        AnsweredMaset ans_mas(maset, questions[i]);
        // Do not store answers for hopeless questions.
        if (ans_mas.max_size() <= upper_size) {
          answered_masets.push_back(ans_mas);
        }
      }
      sort(answered_masets.begin(), answered_masets.end(), CompareAnsweredMasets);
      if (qtradeoff > 0 && answered_masets.size() > qtradeoff) {
        answered_masets.erase(answered_masets.begin() + qtradeoff, answered_masets.end());
      }
    }

    size_t size() const {
      return answered_masets.size();
    }

    AnsweredMaset const & GetAnsweredMaset(size_t i) const {
      return answered_masets[i];
    }

  private:
    vector<AnsweredMaset> answered_masets;

    // Answered masets are sorted in the increasing order of the size of the biggest maset.
    static bool CompareAnsweredMasets(AnsweredMaset const &p1, AnsweredMaset const &p2) {
      return p1.max_size() < p2.max_size();
    }
};

class ABgameBounds {
  public:
    ABgameBounds() {
      lower_bound_table[0] = 0;
      for (int i = 0; i < questions; ++i)
        lower_bound_table[i + 1] = (possible_answers - 1) * lower_bound_table[i] + 1;
    }

    // The information-theoretic lower bound. The maximal size of the maset,
    // which could be solved in question_number questions.
    size_t Lower(int question_number) const {
      return lower_bound_table[question_number];
    }

    // The upper bound is the maximal size of the maset, for which we can
    // prove that it is solvable in question_number questions.
    size_t Upper(int question_number) const {
      return question_number;
    }

  private:
    size_t lower_bound_table[questions + 1];
};

/** Solution search **/

static ABgameBounds const Bounds;
static Maset const *pAllQuestions;
static Maset const *pAllSecrets;

class QuestionGeneratorWrapper {
  public:
    QuestionGeneratorWrapper(Maset const &prev_asked) {
      int prev_asked_size = prev_asked.size();
      if (FUNCTION <= 9 && prev_asked_size < fixed_questions)
        qm = new ABgameFixedQuestion(prev_asked_size);
      else if ((FUNCTION == 10 || FUNCTION > 200) && prev_asked_size < fixed_questions)
        qm = new ABBgameFixedQuestion(prev_asked_size);
      else if (FUNCTION == 11 && prev_asked_size < fixed_questions)
        qm = new ABBgameSemiFixedQuestions();
      else if (prev_asked_size < fixed_questions + itests) {
        // Use nauty to exclude isomorphic questions.
        qm = NautyQuestionGenerator(*pAllQuestions, prev_asked);
      }
      else {
        // Check all possible questions exhaustively.
        qm = pAllQuestions;
      }
      if (SUBCASE1 > 0 && prev_asked_size == 0) {
        // First question
        if (SUBCASE1 <= qm->size()) {
          cout << "first question (#" << SUBCASE1 << " of total "
               << qm->size() << ") = " <<  (*qm)[SUBCASE1 - 1] << endl;
          Maset *tmp = new Maset();
          tmp->push((*qm)[SUBCASE1 - 1]);
          swap(qm, (Maset const *&)tmp);
          if (tmp != pAllQuestions)
            delete tmp;
        }
        else {
          cout << "There is no first question #" << SUBCASE1 <<  "." << endl;
          exit(1);
        }
      }
      else if (SUBCASE1 > 0 && SUBCASE2 > 0 && prev_asked_size == 1) {
        // Second question
        if (SUBCASE2 <= qm->size()) {
          cout << "second question (#" << SUBCASE2 << " of total "
               << qm->size() << ") = " <<  (*qm)[SUBCASE2 - 1] << endl;
          Maset *tmp = new Maset();
          tmp->push((*qm)[SUBCASE2 - 1]);
          swap(qm, (Maset const *&)tmp);
          if (tmp != pAllQuestions)
            delete tmp;
        }
        else {
          cout << "There is no second question #" << SUBCASE2 <<  "." << endl;
          exit(1);
        }
      }
    }

    ~QuestionGeneratorWrapper() {
      if (qm != pAllQuestions)
        delete qm;
    }

    Maset const & GetQuestions() {
      return *qm;
    }

  private:
    Maset const *qm;
};

typedef vector<Answer> AnswerVector;

static bool solvable(Maset const  &maset,
                     Maset const  &question_set,
                     Maset        &prev_asked,
                     AnswerVector &prev_answers,
                     int          remaining_questions) {
  remaining_questions--;
  AllAnswers ans(maset, question_set, Bounds.Lower(remaining_questions));
  for (size_t i = 0; i < ans.size(); ++i) {
    prev_asked.push(ans.GetAnsweredMaset(i).GetAskedQuestion());
    QuestionGeneratorWrapper qgw(prev_asked);
    for (size_t j = 0; j < ans.GetAnsweredMaset(i).size(); ++j) {
      if (ans.GetAnsweredMaset(i).GetMaset(j).size() <= Bounds.Upper(remaining_questions)) {
        if (show) {
          cout << "----------------------------------------\n";
          if (prev_asked.size() > 1) {
            cout << "Former questions and answers\n";
            for (size_t k = 0; k < prev_asked.size() - 1; ++k)
              cout << prev_asked[k] << ' ' << prev_answers[k] << '\n';
            cout << '\n';
          }
          cout << ans.GetAnsweredMaset(i);
        }
        prev_asked.pop();
        return true;
      }
      const Answer &ma = ans.GetAnsweredMaset(i).GetAnswer(j);
      if (FUNCTION == 0) { // AB-game heuristics
        if (prev_answers.size() < (size_t)fixed_questions &&
            prev_answers.size() > 0 &&
            ma.Code() > prev_answers.back().Code()) {
          continue; // Consider only ordered answers to fixed questions.
        }
      }
      else if (FUNCTION >= 201 && FUNCTION <= 299) {
        // the upper bound for 2 pegs and AB-game without white-peg answers heuristics
        size_t s = prev_answers.size();
        if (s < (size_t)fixed_questions && ma.B() != upper_bound_function[FUNCTION - 201][s].answer)
          continue;
      }
      else if (FUNCTION >= 301 && FUNCTION <= 399) {
        // the upper bound for 3 pegs and AB-game without white-peg answers heuristics
        size_t s = prev_answers.size();
        if (s < (size_t)fixed_questions && ma.B() != upper_bound_function[FUNCTION - 301][s].answer)
          continue;
      }
      else if (FUNCTION >= 4001 && FUNCTION <= 4999) {
        // the upper bound for 4 pegs and AB-game without white-peg answers heuristics
        size_t s = prev_answers.size();
        if (s < (size_t)fixed_questions && ma.B() != upper_bound_function[FUNCTION - 4001][s].answer)
          continue;
      }
      prev_answers.push_back(ma);
      bool res = solvable(ans.GetAnsweredMaset(i).GetMaset(j), qgw.GetQuestions(),
                          prev_asked, prev_answers, remaining_questions);
      prev_answers.pop_back();
      if (!res)
        break;
    }
    prev_asked.pop();
  }
  return false;
}

int main() {
  Timer t;

  cout << "function           = " << FUNCTION << '\n'
       << "pegs               = " << pegs << '\n'
       << "colors             = " << colors << '\n'
       << "additional colors  = " << add_colors << '\n'
       << "max questions      = " << questions << '\n'
       << "fixed questions    = " << fixed_questions << '\n'
       << "isomorphism tests  = " << itests << '\n'
       << "question trade off = " << qtradeoff << '\n'
       << "answer trade off   = " << atradeoff << endl;

  if (FUNCTION <= 19 || (FUNCTION > 200 && add_colors == 0)) {
    // AB-game with or without white-peg answers
    pAllSecrets   = new AllABgameConfigurations(0);
    pAllQuestions = new AllABgameConfigurations(add_colors);
  }
  else if (FUNCTION >= 31 && FUNCTION <= 49 && add_colors == 0) {
    // the endgame (the lower bound) for AB-game without white-peg answers
    vector<Peg> vv[pegs];

    if (FUNCTION == 31 && pegs == 3 && colors == 6) { // state A_1
      static const Peg p1[] = {0, 1, 2, 3, 4};
      static const Peg p2[] = {0, 1, 2, 3, 4};
      static const Peg p3[] = {0, 1, 2, 3, 4};
      vv[0] = vector<Peg>(p1, p1 + 5);
      vv[1] = vector<Peg>(p2, p2 + 5);
      vv[2] = vector<Peg>(p3, p3 + 5);
    }
    else if (FUNCTION == 32 && pegs == 3 && colors == 7) { // state A_2
      static const Peg p1[] = {0, 1, 2, 3, 4   };
      static const Peg p2[] = {0, 1, 2, 3,    5};
      static const Peg p3[] = {0, 1, 2,    4, 5};
      vv[0] = vector<Peg>(p1, p1 + 5);
      vv[1] = vector<Peg>(p2, p2 + 5);
      vv[2] = vector<Peg>(p3, p3 + 5);
    }
    else if (FUNCTION == 33 && pegs == 3 && colors == 8) { // state A_3
      static const Peg p1[] = {0, 1, 2, 3, 4,     };
      static const Peg p2[] = {0, 1, 2,       5, 6};
      static const Peg p3[] = {0,       3, 4, 5, 6};
      vv[0] = vector<Peg>(p1, p1 + 5);
      vv[1] = vector<Peg>(p2, p2 + 5);
      vv[2] = vector<Peg>(p3, p3 + 5);
    }
    else if (FUNCTION == 41 && pegs == 4 && colors == 6) { // state B_1
      static const Peg p1[] = {0, 1, 2, 3, 4};
      static const Peg p2[] = {0, 1, 2, 3, 4};
      static const Peg p3[] = {0, 1, 2, 3, 4};
      static const Peg p4[] = {0, 1, 2, 3, 4};
      vv[0] = vector<Peg>(p1, p1 + 5);
      vv[1] = vector<Peg>(p2, p2 + 5);
      vv[2] = vector<Peg>(p3, p3 + 5);
      vv[3] = vector<Peg>(p4, p4 + 5);
    }
    else if (FUNCTION == 42 && pegs == 4 && colors == 7) { // state B_2
      static const Peg p1[] = {0, 1, 2, 3, 4   };
      static const Peg p2[] = {0, 1, 2, 3, 4   };
      static const Peg p3[] = {0, 1, 2, 3,    5};
      static const Peg p4[] = {0, 1, 2,    4, 5};
      vv[0] = vector<Peg>(p1, p1 + 5);
      vv[1] = vector<Peg>(p2, p2 + 5);
      vv[2] = vector<Peg>(p3, p3 + 5);
      vv[3] = vector<Peg>(p4, p4 + 5);
    }
    else if (FUNCTION == 43 && pegs == 4 && colors == 8) { // state B_3
      static const Peg p1[] = {0, 1, 2, 3, 4      };
      static const Peg p2[] = {0, 1, 2, 3,    5   };
      static const Peg p3[] = {0, 1, 2, 3,       6};
      static const Peg p4[] = {0, 1,       4, 5, 6};
      vv[0] = vector<Peg>(p1, p1 + 5);
      vv[1] = vector<Peg>(p2, p2 + 5);
      vv[2] = vector<Peg>(p3, p3 + 5);
      vv[3] = vector<Peg>(p4, p4 + 5);
    }
    else if (FUNCTION == 44 && pegs == 4 && colors == 7) { // state B_4
      static const Peg p1[] = {0, 1, 2, 3, 4   };
      static const Peg p2[] = {0, 1, 2, 3,    5};
      static const Peg p3[] = {0, 1, 2,    4, 5};
      static const Peg p4[] = {0, 1,    3, 4, 5};
      vv[0] = vector<Peg>(p1, p1 + 5);
      vv[1] = vector<Peg>(p2, p2 + 5);
      vv[2] = vector<Peg>(p3, p3 + 5);
      vv[3] = vector<Peg>(p4, p4 + 5);
    }
    else if (FUNCTION == 45 && pegs == 4 && colors == 8) { // state B_5
      static const Peg p1[] = {0, 1, 2, 3, 4      };
      static const Peg p2[] = {0, 1, 2, 3, 4      };
      static const Peg p3[] = {0, 1, 2,       5, 6};
      static const Peg p4[] = {0,       3, 4, 5, 6};
      vv[0] = vector<Peg>(p1, p1 + 5);
      vv[1] = vector<Peg>(p2, p2 + 5);
      vv[2] = vector<Peg>(p3, p3 + 5);
      vv[3] = vector<Peg>(p4, p4 + 5);
    }
    else if (FUNCTION == 46 && pegs == 4 && colors == 8) { // state B_6
      static const Peg p1[] = {0, 1, 2, 3, 4      };
      static const Peg p2[] = {0, 1, 2, 3,    5   };
      static const Peg p3[] = {0, 1, 2,    4,    6};
      static const Peg p4[] = {0,       3, 4, 5, 6};
      vv[0] = vector<Peg>(p1, p1 + 5);
      vv[1] = vector<Peg>(p2, p2 + 5);
      vv[2] = vector<Peg>(p3, p3 + 5);
      vv[3] = vector<Peg>(p4, p4 + 5);
    }
    else if (FUNCTION == 47 && pegs == 4 && colors == 9) { // state B_7
      static const Peg p1[] = {0, 1, 2, 3, 4         };
      static const Peg p2[] = {0, 1, 2, 3,    5      };
      static const Peg p3[] = {0, 1, 2,          6, 7};
      static const Peg p4[] = {         3, 4, 5, 6, 7};
      vv[0] = vector<Peg>(p1, p1 + 5);
      vv[1] = vector<Peg>(p2, p2 + 5);
      vv[2] = vector<Peg>(p3, p3 + 5);
      vv[3] = vector<Peg>(p4, p4 + 5);
    }
    else if (FUNCTION == 48 && pegs == 4 && colors == 8) { // state B_8
      static const Peg p1[] = {0, 1, 2, 3, 4      };
      static const Peg p2[] = {0, 1, 2, 3,    5   };
      static const Peg p3[] = {0, 1,       4, 5, 6};
      static const Peg p4[] = {      2, 3, 4, 5, 6};
      vv[0] = vector<Peg>(p1, p1 + 5);
      vv[1] = vector<Peg>(p2, p2 + 5);
      vv[2] = vector<Peg>(p3, p3 + 5);
      vv[3] = vector<Peg>(p4, p4 + 5);
    }
    else {
      cerr << "ERROR: Wrong parameter combination in main" << endl;
      return -1;
    }

    pAllSecrets   = new AllABBendgameConfigurations(vv);
    pAllQuestions = new AllMastermindConfigurations(0);
  }
  else {
    cerr << "ERROR: Wrong parameter combination in main" << endl;
    return -1;
  }

  bool result;
  if (pAllSecrets->size() <= Bounds.Upper(questions))
    result = true;
  else if (pAllSecrets->size() > Bounds.Lower(questions))
    result = false;
  else {
    Maset prev_asked(questions);
    AnswerVector prev_answers;
    prev_answers.reserve(questions);
    QuestionGeneratorWrapper qgw(prev_asked);
    result = solvable(*pAllSecrets, qgw.GetQuestions(),
                      prev_asked, prev_answers, questions);
  }

  cout << game_name << "(...) "
       << (result ? "<= " : "> ")
       << questions << '\n'
       << "computation time "
       << t.getTime()
       << '\n'
       << endl;

  return result;
}
