#ifndef _ABGAME_H_
#define _ABGAME_H_  1

#include <vector>

#if COLORS < PEGS
  #error: AB-game needs c >= p
#endif

#if ADD_COLORS >= PEGS
  #error: AB-game needs a < p
#endif

// The number of pegs
static const int pegs = PEGS;
// The number of colors
static const int colors = COLORS;
// The maximal repetition count of the additional color in question
static const int add_colors = ADD_COLORS;
// The maximal number of questions
static const int questions = QUESTIONS;
// The number fixed questions at beginning
static const int fixed_questions = FIXED_QUESTIONS;
// The number of question tested for isomorphism
static const int itests = ITESTS;
// The maximum number of considered questions, zero means all questions
static const size_t qtradeoff = QTRADEOFF;
// The maximum number of considered answers, zero means all answers
static const size_t atradeoff = ATRADEOFF;
// Show found strategy
static const int show = SHOW;

typedef char Peg;

class Configuration {
  public:
    Configuration() {
    }

    Configuration(Peg const *p) {
      for (int i = 0; i < pegs; ++i)
        peg_tbl[i] = p[i];
    }

    Peg const & operator[](int i) const {
      return peg_tbl[i];
    }

    void swap(Configuration &c) {
      Peg tmp[pegs];
      memcpy(tmp, peg_tbl, pegs * sizeof(Peg));
      memcpy(peg_tbl, c.peg_tbl, pegs * sizeof(Peg));
      memcpy(c.peg_tbl, tmp, pegs * sizeof(Peg));
    }

  private:
    Peg peg_tbl[pegs];
};

class Maset {
  public:
    Maset() {
    }

    Maset(size_t s) {
      m.reserve(s);
    }

    virtual ~Maset() {
    }

    Configuration const & operator[](size_t i) const {
      return m[i];
    }

    size_t size() const {
      return m.size();
    }

    void push(Configuration const &conf) {
      m.push_back(conf);
    }

    void pop() {
      m.pop_back();
    }

    void swap(Maset &mas) {
      m.swap(mas.m);
    }

  protected:
    ::std::vector<Configuration> m;
};

#endif
