#include <iostream>
#include <cassert>
#include <cstdlib>

using namespace std;

int main(int argc, char *argv[])
{
  if (argc < 2)
  {
    cerr << "The end-game strategy generator for 4 pegs.\n"
            "Usage:\n" << argv[0] << " x\n"
            "Where x is the number of fixed questions.\n";
    exit(1);
  }

  const int x = atoi(argv[1]);
  const int qnum = 4 + 6*(x-1) + 2*(x-1)*(x-2) + (x-1)*(x-2)*(x-3)/6;
  int s, t, u;

  int **qmat = new int*[qnum];
  for (s = 0; s < qnum; ++s)
    qmat[s] = new int[x];

  for (s = 0; s < qnum; ++s)
    for (t = 0; t < x; ++t)
      qmat[s][t] = 0;

  int count = 0;

  qmat[++count][0] = 1;
  qmat[++count][0] = 2;
  qmat[++count][0] = 3;

  for (s = 1; s < x; ++s)
  {
    qmat[++count][0] = 1;
    qmat[count][s]   = 1;
  }

  for (s = 1; s < x; ++s)
  {
    qmat[++count][0] = 1;
    qmat[count][s]   = 2;
  }

  for (s = 1; s < x; ++s)
  {
    qmat[++count][0] = 1;
    qmat[count][s]   = 3;
  }

  for (s = 1; s < x; ++s)
  {
    qmat[++count][0] = 2;
    qmat[count][s]   = 1;
  }

  for (s = 1; s < x; ++s)
  {
    qmat[++count][0] = 2;
    qmat[count][s]   = 2;
  }

  for (s = 1; s < x; ++s)
  {
    qmat[++count][0] = 3;
    qmat[count][s]   = 1;
  }

  for (s = 1; s < x; ++s)
  {
    for (t = s+1; t < x; ++t)
    {
      qmat[++count][0] = 1;
      qmat[count][s]   = 1;
      qmat[count][t]   = 1;
    }
  }

  for (s = 1; s < x; ++s)
  {
    for (t = s+1; t < x; ++t)
    {
      qmat[++count][0] = 1;
      qmat[count][s]   = 1;
      qmat[count][t]   = 2;
    }
  }

  for (s = 1; s < x; ++s)
  {
    for (t = s+1; t < x; ++t)
    {
      qmat[++count][0] = 1;
      qmat[count][s]   = 2;
      qmat[count][t]   = 1;
    }
  }

  for (s = 1; s < x; ++s)
  {
    for (t = s+1; t < x; ++t)
    {
      qmat[++count][0] = 2;
      qmat[count][s]   = 1;
      qmat[count][t]   = 1;
    }
  }

  for (s = 1; s < x; ++s)
  {
    for (t = s+1; t < x; ++t)
    {
      for (u = t+1; u < x; ++u)
      {
        qmat[++count][0] = 1;
        qmat[count][s]   = 1;
        qmat[count][t]   = 1;
        qmat[count][u]   = 1;
      }
    }
  }

  assert(++count == qnum);

  cout << "static const upper_bound_function_t "
       << "upper_bound_function[" << qnum << "][" << x << "] = {" << endl;
  for (s = 0; s < qnum; ++s)
  {
    cout << "  {";
    for (t = 0; t < x; ++t)
    {
      cout << "{" << t << ", " << qmat[s][t] << "}";
      if (t != x-1)
        cout << ", ";
    }
    cout << "}";
    if (s == qnum-1)
      cout << " ";
    else
       cout << ",";
    cout << " // ";
    cout.width(qnum <= 9 ? 1 : qnum <= 99 ? 2 : 3);
    cout << s+1 << endl;
  }
  cout << "};" << endl;

  for (s = 0; s < qnum; ++s)
    delete[] qmat[s];
  delete[] qmat;
}
